/*

The Net Usage Project
http://netusage.mozdev.org/

Copyright (C) 2006-2010 Loune Lam <lpgcritter at nasquan com>
This extension can be freely distributed under
the terms of the Mozilla Public Licence 1.1
See the licence at http://www.mozilla.org/MPL/

---------
Net Usage Item
XPCOM

*/

nsNetUsageAbout.prototype = {
	classDescription: "Net Usage About Helper",
	classID: Components.ID("{e4ea190c-e739-4cc5-9bd1-8ed85c69c685}"),
	contractID: "@mozilla.org/network/protocol/about;1?what=netusage",

	parseQueryString: function (qs) {
		var result = { };
		if (qs == undefined || qs == null)
			return result;
		if (qs[0] == '?')
			qs = qs.substring(1);
		var qa = qs.split('&');
		for (var i in qa) {
			var qai = qa[i].split('=');
			result[qai[0]] = qai[1];
			// TODO look for keys with same key name
		}
		return result;
	},

	firstKey: function (obj) {
		for (k in obj)
			return k;
	},

	getAccounts: function () {
			var dbconn = nuDBConnect();
			
			var periods = { };
			var statement = dbconn.createStatement("SELECT * FROM netusage_accounts");
			while (statement.executeStep()) {
				periods[statement.getInt32(0)] = statement.getUTF8String(1)+" - "+statement.getUTF8String(2);
			}
			return periods;
	},
	
	getPeriods: function () {
			var dbconn = nuDBConnect();
			
			var periods = { };
			var statement = dbconn.createStatement("SELECT * FROM netusage_periods ORDER BY startdate DESC");
			while (statement.executeStep()) {
				periods[statement.getInt32(0)] = [statement.getInt32(1), statement.getInt32(2)];
			}
			return periods;
	},															

	objToDropDown: function (id, oc, obj, sel, type) {
		var s = "<select id='"+id+"' onchange='"+oc+"'>";
		for (var k in obj) {
			var selected = (sel == k) ? " selected='selected' " : "";
			if (type == 'periods')
				s += "<option value='"+k+"'"+selected+">"+(new Date(obj[k][0]*1000).toLocaleFormat("%a %d %b %Y")) +" - "+ (new Date(obj[k][1]*1000).toLocaleFormat("%a %d %b %Y"))+"</option>";
			else
				s += "<option value='"+k+"'"+selected+">"+obj[k]+"</option>";
				
		}
		s += "</select>";
		return s;
	},

	getUsageDetails: function (format, account, period) {
		var s = "";
		
		var dbconn = nuDBConnect();
		
		var types = { };
		var statement = dbconn.createStatement("SELECT * FROM netusage_types");
		while (statement.executeStep()) {
			types[statement.getInt32(0)] = statement.getUTF8String(1);
		}
		
		var statement = dbconn.createStatement("SELECT * FROM netusage_data WHERE account_id = ?1 AND period_id = ?2 ORDER BY date DESC, type_id ASC");
		
		statement.bindInt32Parameter(0, account);
		statement.bindInt32Parameter(1, period);
		var i = -1;
		var lastdate = 0;
		if (format == 'table') {
			var header = "<th>Date Time</th>";
			s = "";
			while (statement.executeStep()) {
				var date = statement.getInt32(3);
				if (date != lastdate) {
					s+="</tr><tr><td style='text-align: right'>"+(new Date(date*1000).toLocaleFormat("%d-%b-%y %H:%M"))+"</td>";
					i++;
				}
				if (i == 0) {
					header += "<th>"+types[statement.getInt32(2)]+"</th>";
				}
				lastdate = date;
				s += "<td style='text-align: right'>"+statement.getInt32(4)+"</td>";
			}
			s = "<table border=1><tr>"+header+"</tr><tr>"+s+"</tr></table>";
		}
		return s;
	},


	//
	// nsIAboutModule
	//

	newChannel: function(aURI) {
		// process URI
		var uri = decodeURI(aURI.spec);
		var uri = uri.split(/\?/);
		var querystr = uri[1];
		var s = '';
		
		var scriptinc = Components.classes["@mozilla.org/moz/jssubscript-loader;1"].getService().QueryInterface(Components.interfaces.mozIJSSubScriptLoader);
		if (typeof(__USAGELIB_JS__) == 'undefined')
			scriptinc.loadSubScript("chrome://netusageitem/content/usagelib.js");
		
		if (querystr == 'add') {
			// open new chrome channel redir
			var iosvc = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
			var chan = iosvc.newChannel("chrome://netusageitem/content/add.xhtml", null, null);
		
		} else if (querystr == 'theme') {
			// open new chrome channel redir
			var iosvc = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
			var chan = iosvc.newChannel("chrome://netusageitem/content/theme.html", null, null);
		
		} else {

			// dynamic content
			s += '<html><head><title>About your net usage</title></head><body>';

			s += '<h1>My Net Usage</h1>';
			
			var periods = this.getPeriods();
			var accounts = this.getAccounts();
			var qs = this.parseQueryString(querystr);
			if (qs['a'] == undefined)
				qs['a'] = this.firstKey(accounts);
			if (qs['p'] == undefined)
				qs['p'] = this.firstKey(periods);
				
			var ddjs = 'window.location = window.location.toString().substring(0, window.location.toString().indexOf("?") != -1 ? window.location.toString().indexOf("?") : window.location.toString().length) + "?a="+document.getElementById("accounts").value+"&p="+document.getElementById("periods").value ';
			s += (this.objToDropDown('accounts', ddjs, accounts, parseInt(qs['a']), ''));
			s += (this.objToDropDown('periods', ddjs, periods, parseInt(qs['p']), 'periods'));

			s += (this.getUsageDetails('table', qs['a'], qs['p']));

			s += '</body></html>';

			// string stream
			var strm = Components.classes['@mozilla.org/io/string-input-stream;1'].createInstance(Components.interfaces.nsIStringInputStream);
			strm.setData(s, -1);

			// create new channel
			var chan = Components.classes["@mozilla.org/network/input-stream-channel;1"].createInstance(Components.interfaces.nsIInputStreamChannel);
			chan.QueryInterface(Components.interfaces.nsIChannel);
			chan.setURI(aURI);
			chan.contentStream = strm;
			chan.contentType = "text/html";
		}
		
		return chan;
	},
	
	getURIFlags: function(aURI) {
		return Components.interfaces.nsIAboutModule.URI_SAFE_FOR_UNTRUSTED_CONTENT | Components.interfaces.nsIAboutModule.ALLOW_SCRIPT;
	},


	//
	// nsISupports
	//


	QueryInterface: function (iid) {
		if (
			!iid.equals(Components.interfaces.nsISupports) &&
			!iid.equals(Components.interfaces.nsIAboutModule)) {
			throw Components.results.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},

}



function nsNetUsageAbout() {


}




// MODULE
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function NSGetModule(compMgr, fileSpec) {
	return XPCOMUtils.generateModule([nsNetUsageAbout]);
}

if (XPCOMUtils.generateNSGetFactory)
	var NSGetFactory = XPCOMUtils.generateNSGetFactory([nsNetUsageAbout]);
