/*

The NetUsage Project
http://netusage.mozdev.org/

Copyright (C) 2006-2010 Loune Lam <lpgcritter at nasquan com>
This extension can be freely distributed under
the terms of the Mozilla Public Licence 1.1
See the licence at http://www.mozilla.org/MPL/

---------
Net Usage Item
XPCOM

*/

nsNetUsageItem.prototype = {

	classDescription: "Net Usage Item",
	classID: Components.ID("{489e3485-7a4a-4438-9185-bf14c0c317ba}"),
	contractID: "@netusage.iau5.com/netusageitem;1",

	//
	// nsINetUsageItem
	//

	get itemid() { return this.mitemid; },
	get status() { return this.mStatus; },
	get errorText() { return this.mErrorText; },
	get lastFetched() { return this.mLastFetched; },
	get ispid() { return this.isp ? this.isp.id : ''; },
	getVar: function(k) { return this.vars == null ? null : this.vars[k]; },
	setVar: function(k, val) { if (this.vars != null) this.vars[k] = val; },

	//
	// private
	//

	init: function(aitemid) {
		// includes
		var scriptinc = Components.classes["@mozilla.org/moz/jssubscript-loader;1"].getService().QueryInterface(Components.interfaces.mozIJSSubScriptLoader);
		
		if (typeof(__USAGELIB_JS__) == 'undefined')
			scriptinc.loadSubScript("chrome://netusageitem/content/usagelib.js");
		this.getISPInterface = nuGetISPInterface;

		this.observerSvc = Components.classes["@mozilla.org/observer-service;1"].getService().QueryInterface(Components.interfaces.nsIObserverService);

		nuDump('XXXX init: '+aitemid);

		this.mitemid = aitemid;
		this.checkInterval = -1;
		this.isp = null;
		this.mLastFetched = 0;
		
		this.mDBConn = nuDBConnect();
		
		this.timer = Components.classes['@mozilla.org/timer;1'].createInstance(Components.interfaces.nsITimer);
		this.lockedtimer = Components.classes['@mozilla.org/timer;1'].createInstance(Components.interfaces.nsITimer);
		
		var prefSvc = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
		//var pref = prefSvc.getBranch("extensions.netusage."+this.mitemid+".");
		var pref = prefSvc.getBranch(""); // TODO why doesn't above work?
		pref.QueryInterface(Components.interfaces.nsIPrefBranch2);
		pref.addObserver("extensions.netusage."+this.mitemid+".", this, false);
		nuDump("extensions.netusage."+this.mitemid+".");
		
		
		// load prefs should trigger initTimer()
		this.loadPrefs();
	},
	
	uninit: function() {
		var prefSvc = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
		var pref = prefSvc.getBranch("");
		pref.QueryInterface(Components.interfaces.nsIPrefBranch2);
		pref.removeObserver("extensions.netusage."+this.mitemid+".", this);
		this.timer.cancel();
		this.lockedtimer.cancel();
		this.abort();
		//this.observerSvc.removeObserver(this, '');
		this.observerSvc = null;
		this.timer = null;
		nuDump('UnInit !!!!!!!!!!!!!!!!');
	},

	// initialises timer
	initTimer: function () {
		this.timer.cancel();
		if (this.isp != null && this.checkInterval >= 600)
			this.timer.init(this, this.checkInterval * 1000, this.timer.TYPE_REPEATING_SLACK);
		else
			nuDump('timer disabled');
	},

	//
	// nsINetUsageItem
	//
	
	setISPInterface: function(objjson) {
		//this.isp = eval('(' + objjson + ')');
	},
	
	getVarsObjStr: function() {
		return nuDumpobjstr(this.vars);
	},
	
	getVarExpand: function(s) {
		return nuVarExpand(s, this.vars, false, false, false, this.isp, 
			this.filterFuncList, this.filterFunc);
	},
	
	abort: function() {
		if (this.mStatus == 'fetching') {
			// abort
			nuDump('aborting...');
			this.xmlhttp.abort();
			this.endUpdate('aborted', 'User aborted usage update');
		}
		
		else if (this.mStatus == 'fetching_queued') {
			// abort while still stuck in lock queue
			nuDump('queue aborting...');
			this.lockedtimer.cancel();
			
			this.mStatus = "aborted";
			this.mErrorText = "User aborted usage update";
			this.observerSvc.notifyObservers(this, 'netusage-update-gui', 'draw');
		}
	},
	
	endUpdate: function(status, errorText) {
		nuDump('Fetch ended');
		
		var nusvc = Components.classes["@netusage.iau5.com/netusagesvc;1"].getService(Components.interfaces.nsINetUsageSvc);
		nusvc.releaseLock("ISP_" + this.isp.name);
			
		if (status != null)
			this.mStatus = status;
		if (errorText != null)
			this.mErrorText = errorText;
		
		// write or delete debug file?
		if (this.mDebug || this.status == 'error-process' || this.status == 'error-parse') {
			this.debugDump(this.mErrorText != '' ? this.mErrorText+"\n" : "Debug mode\n");
		} else {
			var file = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties).get("ProfD", Components.interfaces.nsIFile);
			file.append("netusage_debug.txt");
			if (file.exists())
				file.remove(false);
		}
		
		try {
		this.observerSvc.removeObserver(this, 'http-on-examine-response');
		} catch(e) {}
		if (this.mOpenPageCmd == null)
			this.observerSvc.notifyObservers(this, 'netusage-update-gui', 'draw');
		else
			this.observerSvc.notifyObservers(this, 'netusage-open-page', this.mOpenPageCmd);

		this.xmlhttp = null;
		this.mDebughttp = '';
	},

	updateUsage: function(mod) {
		try {
		
		this.lockedtimer.cancel();

		if (this.isp == null) {
			nuDump('updateUsage this.isp == null');
			this.mStatus = "error";
			this.mErrorText = 'Please select an ISP in your preferences.';
			return false;
		}
		
		if ((mod & Components.interfaces.nsINetUsageItem.UPDATE_STARTUP) && this.mFetchOnStartup == 0) {
			this.mStatus = "error"; // TODO make warning
			this.mErrorText = 'Fetch On Startup disabled. Please click to refresh.';
			return false;
		}
		
		if (((mod & Components.interfaces.nsINetUsageItem.UPDATE_STARTUP) || (mod & Components.interfaces.nsINetUsageItem.UPDATE_AUTO))
			&& nuMasterPasswordStatus() == Components.interfaces.nsIPKCS11Slot.SLOT_NOT_LOGGED_IN && this.mFetchOnStartup != 1) {
			nuDump('master password set and not logged in');
			this.mStatus = "error";
			this.mErrorText = 'Master password is set. Please click to refresh.';
			return false;
		}
		
		if (this.mStatus == 'fetching') {
			// we're already fetching!
			return false;
		}
		
		// make sure another nui with same ISP isn't running, due to conflicts involving cookies
		var nusvc = Components.classes["@netusage.iau5.com/netusagesvc;1"].getService(Components.interfaces.nsINetUsageSvc);
		if (!nusvc.acquireLock("ISP_" + this.isp.name)) {
			nuDump("FETCHING QUEUED of " + this.isp.name);
			this.mStatus = 'fetching_queued';
			this.lockedtimer.init(this, 2000, this.timer.TYPE_ONE_SHOT);
			return true;
		}
	
		this.mDebughttp = '';
		this.mStatus = 'fetching';
		// beyond this point, we will always return true and we will always notify to show error //
		
		// update all client displays (so they flash)
		this.observerSvc.notifyObservers(this, 'netusage-update-gui', 'draw');

		// clean step, vars
		this.step = 0;
		this.nextstep = 0;
		this.mVarMapped = false;
		this.mErrorText = "";
		this.mOpenPageCmd = null;
		this.mUsagePageStep = -1;
		if (mod & Components.interfaces.nsINetUsageItem.UPDATE_OPEN_PAGE)
			this.prepareUsagePageOpen();
		
		this.mOldVars = this.vars;
		this.vars = new Array();
		this.vars['ispid'] = this.isp.id;
		this.vars['ispname'] = this.isp.name;
		if (this.isp.vars) {
			// copy over isp vars
			for (i in this.isp.vars)
				this.vars[i] = this.isp.vars[i];
		}

		// fetch username and password
		var ublogin;
		if (!(ublogin = nuLoadPass('net:usage:bar:'+this.mitemid))) {
			this.endUpdate('error-up', "Username and Password required");
			return true;
		}
		this.vars['USERNAME'] = ublogin[1];
		this.vars['PASSWORD'] = ublogin[2];

		// TODO set this here or when we finish?
		if (!(mod & Components.interfaces.nsINetUsageItem.UPDATE_OPEN_PAGE))
			this.mLastFetched = Math.floor((new Date()).getTime()/1000);
		
		// load [override] vars from pref
		nuLoadPrefList("extensions.netusage."+this.mitemid+".vars.", this.vars);
		
		// some predefined vars
		var now = new Date();
		var monthto3l = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'];
		this.vars['now_month'] = now.getMonth()+1;
		this.vars['now_month_3l'] = monthto3l[now.getMonth()];
		this.vars['now_day'] = now.getDate();
		this.vars['now_year'] = now.getFullYear();
		
		// debug
		// nuDumpobj('this.vars', this.vars);
		
		// for nuking WWW-authenticate
		this.observerSvc.addObserver(this, 'http-on-examine-response', false);
		
		// start the process
		this.resumeProcess();
		
		} catch(e) {
			this.endUpdate('error', 'Exception: ' + e + "\n" + e.stack);
		}
		
		return true;
	},
	
	prepareUsagePageOpen: function() {
		var start = 0, usagepage = -1, lastgp = 0, nextisusagepage = false;
		for (var i = 0; i < this.isp.process.length; i++) {
			if (typeof(this.isp.process[i]) == 'string') {
				if (this.isp.process[i] == 'STARTUSAGEPAGE')
					start = i+1;
				else if (this.isp.process[i] == 'USAGEPAGE')
					nextisusagepage = true;
				else if (this.isp.process[i].substring(0, 3) == 'GET' || // GETBASIC POSTBASIC
					this.isp.process[i].substring(0, 4) == 'POST') {
					if (nextisusagepage)
						usagepage = i;
					else
						lastgp = i;
					nextisusagepage = false;
				}
			}
		}
		nuDump("lastgp "+lastgp+" usagepage "+usagepage+" start "+start);
		if (usagepage == -1)
			usagepage = lastgp;
		this.mUsagePageStep = usagepage;
		this.step = start;
		this.nextstep = start;
	},
	
	resumeProcess: function() {
		var gonext = true; // needed because asynchronous XMLHttpRequest breaks out of process loop
		while (gonext) {
			this.vars['laststep'] = this.step;
			this.step = this.nextstep;
			
			gonext = this.stepProcess();
			
			if (this.mStatus != 'fetching') {
				// process step error occured OR all done
				this.endUpdate(null, null);
				break;
			}
			nuDump("resumeProcess done step "+this.step);
		}
		nuDump("resumeProcess end");
	},

	// steps this.isp.process returns true if caller should immediately call next
	stepProcess: function() {
		nuDump("stepProcess() step "+this.step);
		
		// check loop LOOP MAX = 5
		if (typeof(this.vars['step' + this.step + 'count']) == 'undefined') {
			this.vars['step' + this.step + 'count'] = 0;
		}
		this.vars['step' + this.step + 'count']++;
		
		if (this.vars['step' + this.step + 'count'] > 5) {
			nuDump("stepUpdateUsage() step loop maximum exceeded");
			
			this.mStatus = 'error-process';
			this.mErrorText = 'Step looping maximum exceeded!';
			return false;
		}
		
		if (typeof(this.step) == 'string') {
			// resolve label
			for (var i in this.isp.process) {
				if (typeof(this.isp.process[i]) == 'string' && this.isp.process[i] == "LABEL "+this.step) {
					this.step = parseInt(i)+1;
					break;
				}
			}
		}
		if (typeof(this.step) == 'string') { // still eq string, then label not found
			this.mStatus = 'error-process';
			this.mErrorText = 'Unable to find LABEL '+this.step;
			return false;
		}
		
		this.nextstep = this.step + 1;

		// NOW See what step we are up to and what type of step is it
		if (this.step >= this.isp.process.length) {
			nuDump("process step finish");
			// we've arrived at the finish
			// post process
			if (!this.mVarMapped)
				this.mapVariables();
				
			this.mapAuxiliaryVariables();
			
			nuDump("map var finish");
			// update last fetched
			this.mLastFetched = Math.floor((new Date()).getTime()/1000);
			
			// record usage to database
			var account_id = nuDBGetOrCreate(this.mDBConn, "netusage_accounts", { "isp": this.isp.id, "username": this.vars['USERNAME'] });
			var period_id = nuDBGetOrCreate(this.mDBConn, "netusage_periods", {'startdate': this.vars['startdate'], 'enddate': this.vars['enddate']});
			for (var i in this.vars) {
				if (i.indexOf('-quota') != -1 || i == 'startdate' || i == 'enddate' || i == 'rolldate')
					continue; // quotas, dates aren't recorded this way
				if (typeof(this.vars[i]) != 'number')
					continue; // we are recording numbers only
				if (i.indexOf('download') == -1 && i.indexOf('upload') == -1 && i.indexOf('free') == -1 && i.indexOf('usage') == -1)
					continue;
				var quota = (this.vars[i+'-quota'] != undefined) ? this.vars[i+'-quota'] : 0;
				var type_id = nuDBGetOrCreate(this.mDBConn, "netusage_types", { "type_name": i, "quota": quota });
				nuDBInsert(this.mDBConn, "netusage_data", {'account_id': account_id, 'period_id': period_id, 'type_id': type_id, date: this.mLastFetched, 'usage': this.vars[i]});
			}

			
			// debug
			nuDump('this.vars = {');
			for (var i in this.vars) {
				if (i == 'USERNAME' || i == 'PASSWORD')
					continue;
				var zzz = this.vars[i];
				if (typeof(zzz) == 'string') {
					zzz = zzz.substr(0, 100);
				}
				nuDump(i+': '+zzz);
			}
			nuDump('}');

			this.mStatus = 'done';
			return false;
		}
		// process step function
		else if (typeof(this.isp.process[this.step]) == 'function' && typeof(this.isp.process[this.step]['exec']) != 'function') {
			nuDump("process step function");
			
			var ns = (this.isp.process[this.step])(this.vars);
			if (typeof(ns) == 'number' || typeof(ns) == 'string')
				this.nextstep = ns;

			if (this.vars['errortext']) {
				// process function raised an error
				this.mStatus = 'error-isp';
				this.mErrorText = this.vars['errortext'];
				return false;
			}
		}
		// process step command
		else if (typeof(this.isp.process[this.step]) == 'string') {
			nuDump("process step command");
			// substitute all vars in URL cmd string
			var s = nuVarExpand(this.isp.process[this.step], this.vars, true, false, false, this.isp, 
				this.filterFuncList, this.filterFunc);
			
			if (this.mUsagePageStep == this.step) {
				// we arrived at the usage page step, open usage page in GUI
				this.mOpenPageCmd = s;
				this.nextstep = 1000;
				this.mStatus = 'done';
				// restore old vars in case it gets queried
				this.vars = this.mOldVars;
				return false;
			}

			var params = s.split(/ /);
			if (params[0] == 'POST' || params[0] == 'GET' || params[0] == 'POSTBASIC' || params[0] == 'GETBASIC') {
				this.stepUpdateUsage(params);
				return false;
			}
			else if (params[0] == 'MAPVARS')
				this.mapVariables();
			else if (params[0] == 'LABEL')
				{} // do nothing, skip
			else if (params[0] == 'EXIT')
				this.nextstep = 1000;
			else if (params[0] == 'SKIPNEXTSTEP')
				this.nextstep++;
			else if (params[0] == 'DELETECOOKIES')
				nuDeleteCookies(params); // we really should be cutting off the first param, but meh
			else if (params[0] == 'STARTUSAGEPAGE') {
				// all the steps below this is usage page fetching related
				// here we are not fetching usage page, don't proceed
				this.nextstep = 1000;
			}
			else if (params[0] == 'USAGEPAGE') {
				// here, the next action should be notifying the GUI to show the usage page
				// all logic is at prepareUsagePageOpen though
			}
			else if (params[0] == 'REFRESHREDIRECT') {
				// handle http equiv redirects
				var ra = / content=(?:["']|)[0-9]+;url=(?:["']|)([^"']+)/i.exec(this.vars['lasttext']);
				if (ra) {
					this.stepUpdateUsage(['GET', ra[1]]);
					this.nextstep = this.step;
					return false;
				}
			}
			else {
				nuDump('UKNOWN COMMAND '+params[0]);
				this.mStatus = 'error-process';
				this.mErrorText = 'Unknown process command (S'+this.step+'): '+s;
				return false;
			}
		}
		// process step parsing
		else if (typeof(this.isp.process[this.step]) == 'object' || nuIsRegExp(this.isp.process[this.step])) {
			nuDump("process step parsing");
			this.nextstep = this.parse(this.isp.process[this.step]);

			if (this.nextstep == 'ERROR') {
				if (this.mStatus != 'error-parse') {
					if (this.vars[this.getParseNum(this.step) + ":1"]) {
						this.mErrorText = 'ISP Error: '+this.vars[this.getParseNum(this.step) + ":1"] + " (S" + this.step + ")";
					} else {
						this.mErrorText = 'Login Failed. Please check your username and password.';
					}
					this.mStatus = 'error-isp';
				}
				this.nextstep = -1;
				return false;
			}
		}
		
		nuDump('nextstep: '+this.nextstep);
		return true;
		
	},
		
	stepUpdateUsage: function(params) {
		var poststr;

		if (params[0] == 'POST' || params[0] == 'POSTBASIC')
			poststr = params[2];
		else
			poststr = null;

		this.xmlhttp = Components.classes['@mozilla.org/xmlextras/xmlhttprequest;1'].createInstance(Components.interfaces.nsIXMLHttpRequest);

		//this.xmlhttp.QueryInterface(Components.interfaces.nsIDOMEventTarget);
		this.xmlhttp.addEventListener('load', this, false);
		this.xmlhttp.addEventListener('error', this, false);
		
		var url = params[1];
		nuDump('OPEN '+params[0]+';'+url);
		this.xmlhttp.open(params[0] == 'GETBASIC' ? 'GET' : (params[0] == 'POSTBASIC' ? 'POST' : params[0]) , url, true);
		
		var ngd = nuGetDomain(url);
		if (ngd != null)
			nuAddCookieDomain(ngd);
		
		// throw off caching
		this.xmlhttp.setRequestHeader("If-Modified-Since", "Sat, 1 Jan 2000 00:00:00 GMT");
		if (this.vars['NETUSAGE_UA'])
			this.xmlhttp.setRequestHeader("User-Agent", "NetUsage/"+nuGetVersion());
		if (this.vars['referer'])
			this.xmlhttp.setRequestHeader("Referer", this.vars['referer']);
		
		if (params[0] == 'POST' || params[0] == 'POSTBASIC')
			this.xmlhttp.setRequestHeader('Content-type', 'application/x-www-form-urlencoded');
		if (params[0] == 'GETBASIC' || params[0] == 'POSTBASIC')
			this.xmlhttp.setRequestHeader('Authorization', 'Basic ' + nuEncodeBase64(this.vars['USERNAME']+':'+this.vars['PASSWORD']));
		// this.xmlhttp.channel.notificationCallbacks = this; //TODO

		this.vars['referer'] = url;
		this.mDebughttp += this.step+') '+params[0]+' '+url+' '+(poststr != null ? poststr : '')+"\n";
		this.xmlhttp.send(poststr);

	},
	
	debugDump: function(strprefix) {
		// dump response, source into profile
		delete this.vars['lasttext'];
		var doc = strprefix+"Version "+nuGetVersion()+"\n"+this.mDebughttp+"\n\n"+nuDumpobjstr('this.vars', this.vars)+"\nNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNN\n";
		if (this.xmlhttp)
			doc += this.xmlhttp.responseText;
		
		// sanitize transcript to protect u/p
		if (typeof(this.vars['USERNAME']) != 'undefined' && this.vars['USERNAME'] != null && this.vars['USERNAME'] != '') {
			doc = doc.replace(new RegExp(nuEscapeRegexp(encodeURIComponent(this.vars['USERNAME'])), 'g'), 'UXXX');
			doc = doc.replace(new RegExp(nuEscapeRegexp(this.vars['USERNAME']), 'g'), 'UXXX');
		} 
		if (typeof(this.vars['PASSWORD']) != 'undefined' && this.vars['PASSWORD'] != null && this.vars['PASSWORD'] != '') {
			doc = doc.replace(new RegExp(nuEscapeRegexp(encodeURIComponent(this.vars['PASSWORD'])), 'g'), 'PXXX');
			doc = doc.replace(new RegExp(nuEscapeRegexp(this.vars['PASSWORD']), 'g'), 'PXXX');
		}
		
		var file = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties).get("ProfD", Components.interfaces.nsIFile);
		file.append("netusage_debug.txt");
		var fs = Components.classes["@mozilla.org/network/file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
		fs.init(file, 0x02 | 0x08 | 0x20, 0664, 0); // write, create, truncate
		fs.write(doc, doc.length);
		fs.close();
	},
	
	getParseNum: function(step) {
		var parsenum = 0;
		for (var i = 0; i < step; i++) {
			if (typeof(this.isp.process[i]) == 'object' || nuIsRegExp(this.isp.process[i]))
				parsenum++;
		}
		return parsenum;
	},

	// parsing, called by dom event load
	parse: function(ae) {
		var rv, i;
		var branch = 0;
		var parsenum = this.getParseNum(this.step);
		
		// we now check function first because in FF3beta regexp is type object (FF2 is function)
		if (nuIsRegExp(ae)) {
			// only 1 regexp
			nuDump("ae[exec] typeof regexp");
			rv = this.parseRegexp(ae, this.vars['lasttext']);
			this.vars['parse'+parsenum] = 0;
		}
		else if (typeof(ae) == 'object') {
			nuDump("ae typeof object/array");
			
			// 3rd arg is an array... loop through it
			for (i in ae) {
				// theoretically, we can mix xpath expr and regexp expr leafs
				if (nuIsRegExp(ae[i])) {
					// one regexp out of multiple (regexp, xpath)
					nuDump("one regexp out of multiple");
					rv = this.parseRegexp(ae[i], this.vars['lasttext']);
				}
				else if (typeof(ae[i]) == 'object') {
					// one xpath out of multiple (regexp, xpath)
					nuDump("one xpath out of multiple");
					rv = this.parseXML(ae[i], this.xmlhttp.responseXML);
				}
				else if (typeof(ae[i]) == 'string' && ae["XPATHVERIFY"]) {
					// only 1 xpath branch, (one xpath branch is a string array of xp expr)
					nuDump("1 xpath branch");
					rv = this.parseXML(ae, this.xmlhttp.responseXML);
					this.vars['parse'+parsenum] = 0; // the branch we took
					break;
				}
				else if (typeof(ae[i]) == 'string' || typeof(ae[i]) == 'number') {
					// this is a special directive of the path to take if we match
					branch--;
				}
				
				// did we find a match?
				if (rv) {
					this.vars['parse'+parsenum] = branch; // the branch we took
					break;
				}
				branch++;
			}
		}

		if (!rv) {
			// parsing failed
			nuDump("parsing failed");
			this.mStatus = 'error-parse';
			this.mErrorText = 'A parsing error occured (S'+this.step+')';
			return 'ERROR';
		}
		
		// move stuff in rv into this.vars, append 'parsenum:' to key
		for(var j in rv) {
			this.vars[parsenum  + ":" + j] = rv[j];
		}
		
		if (ae["XPATHVERIFY"] != undefined)
			return this.nextstep;
		var ni = parseInt(i) + 1;
		return (typeof(ae[ni]) == 'number' || typeof(ae[ni]) == 'string') ? ae[ni] : this.nextstep;
	},


	// parsing text using regexp
	parseRegexp: function(aExp, aDoc) {
		var a = aExp.exec(aDoc);
		this.vars['last_regexp'] = aExp.toString();
		if (!a) {
			nuDump('parseRegexp failed');
			return null;
		}
		a[0] = 0;
		return a;
	},


	// parsing XML using xpath
	parseXML: function(aPaths, aDoc) {
		var nsr = (typeof(this.vars['NamespaceResolver']) != 'undefined') ? this.vars['NamespaceResolver'] : null;
		try {
			if (aPaths['XPATHVERIFY'] && !aDoc.evaluate(aPaths['XPATHVERIFY'], aDoc, nsr, 9, null).singleNodeValue) {
				nuDump('parse XPATH failed');
				return null;
			}
		} catch(e) {
			nuDump('parse XPATH failed ' + e);
			return null;
		}
		for (var k in aPaths) {
			if (k == 'XPATHVERIFY') continue;
			nuVarPipe(k, aDoc.evaluate(aPaths[k], aDoc, nsr, 2, null).stringValue, this.vars, false, false, true, this.isp, 
				this.filterFuncList, this.filterFunc);
		}
		return [];
	},
	
	// get the start and end dates as an array [start, end]
	getStartEndDate: function(vstart, vend, vroll) {
		var startdate, enddate;
		var now = new Date();
		
		if (vstart)
			startdate = nuStrToDate('start', vstart);
		if (vend)
			enddate = nuStrToDate('end', vend);

		if (vroll) {
			// roll date AKA next start date specified - now same as end date due to [ ) rule
			enddate = nuStrToDate('roll', vroll);
			if (!vstart) {
				// make guess as same day last month
				startdate = new Date(enddate.getFullYear(), enddate.getMonth()-1, enddate.getDate(), 0, 0, 0);
			}
		}
		else if (!vstart && vend) {
			// only end date specified, guess start date
			startdate = new Date(enddate.getFullYear(), enddate.getMonth()-1, enddate.getDate(), 0, 0, 0);
		}
		else if (vstart && !vend) {
			// only start date specified, guess end date
			enddate = new Date(startdate.getFullYear(), startdate.getMonth()+1, startdate.getDate(), 0, 0, 0);
		}
		else if (!vstart && !vend) {
			// calendar month
			startdate = new Date(now.getFullYear(), now.getMonth(), 1, 0, 0, 0);
			enddate = new Date(now.getFullYear(), now.getMonth()+1, 1, 0, 0, 0);
		}
		return [startdate, enddate];
	},
	
	// map these variables
	mapVariables: function() {
		this.mVarMapped = true;
		
		var a;
		// var mapping
		for (var i = 0; i < this.isp.process.length; i++) {
			if (a = this.isp.varsDisplay[i+','+this.vars["parse"+i]])
				for(k in a) {
					if (typeof(a[k]) == 'object') {
						nuVarExpand(a[k][0], this.vars, false, false, true, this.isp, 
							this.filterFuncList, this.filterFunc);
						if (a[k][1] != undefined)
							nuVarExpand(a[k][1], this.vars, false, false, true, this.isp, 
								this.filterFuncList, this.filterFunc);
					} else
						nuVarExpand(a[k], this.vars, false, false, true, this.isp, 
							this.filterFuncList, this.filterFunc);
				}
		}
		
		// filter out zero quotas
		if (this.vars['downloadpo-quota'] != undefined && this.vars['downloadpo-quota'] == 0)
			delete this.vars['downloadpo-quota'];
		if (this.vars['usagepo-quota'] != undefined && this.vars['usagepo-quota'] == 0)
			delete this.vars['usagepo-quota'];
		if (this.vars['usage2-quota'] != undefined && this.vars['usage2-quota'] == 0)
			delete this.vars['usage2-quota'];
		
		// start end date inference
		var d = this.getStartEndDate(this.vars['startdate'], this.vars['enddate'], this.vars['rolldate']);

		this.vars['startdate'] = Math.floor(d[0].getTime()/1000);
		this.vars['enddate'] = Math.floor(d[1].getTime()/1000);
		
		// adjust to ISP timezone
		if (typeof(this.vars['isptz']) != 'undefined' && this.vars['isptz'] != 86400) {
			var nowtz = (new Date()).getTimezoneOffset()*60;
			this.vars['startdate'] -= nowtz + this.vars['isptz'];
			this.vars['enddate'] -= nowtz + this.vars['isptz'];
		}
	},
	
	mapAuxiliaryVariables: function() {
		
		// usage var doesn't exist, default to download var
		if (typeof(this.vars['usage']) == 'undefined')
			this.vars['usage'] = this.vars['download'];
		if (typeof(this.vars['usage-quota']) == 'undefined')
			this.vars['usage-quota'] = this.vars['download-quota'];
		if (typeof(this.vars['usagepo']) == 'undefined' && typeof(this.vars['downloadpo']) != 'undefined')
			this.vars['usagepo'] = this.vars['downloadpo'];
		if (typeof(this.vars['usagepo-quota']) == 'undefined' && typeof(this.vars['downloadpo-quota']) != 'undefined')
			this.vars['usagepo-quota'] = this.vars['downloadpo-quota'];
		
		// TODO we should separate usage2 and usage2-quota (like above) so that if usage2-quota is filled in prefs but the regexp vars are download2, usage2 will never get filled (cause usage2-quota is already filled)
		if (typeof(this.vars['usage2-quota']) == 'undefined' && typeof(this.vars['download2-quota']) != 'undefined') {
			this.vars['usage2'] = this.vars['download2'];
			this.vars['usage2-quota'] = this.vars['download2-quota'];
		}
		// or else
		if (typeof(this.vars['usage2-quota']) == 'undefined' && typeof(this.vars['usagepo-quota']) != 'undefined') {
			this.vars['usage2'] = this.vars['usagepo'];
			this.vars['usage2-quota'] = this.vars['usagepo-quota'];
			this.vars['usage2-type'] = 'offpeak';
		}
	
		// percentages
		if (typeof(this.vars['usage-quota']) != 'undefined')
			this.vars['usagepc'] = parseInt(parseInt(this.vars['usage']) / parseInt(this.vars['usage-quota']) * 100);
		if (typeof(this.vars['usage2-quota']) != 'undefined')
			this.vars['usagepc2'] = parseInt(parseInt(this.vars['usage2']) / parseInt(this.vars['usage2-quota']) * 100);
			
		var cdate = Math.floor((new Date()).getTime()/1000) - parseInt(this.vars['startdate']);
		var mdate = parseInt(this.vars['enddate']) - parseInt(this.vars['startdate']);
		this.vars['datepc'] = parseInt(cdate / mdate * 100);
	},
	
	filterFuncList: ['B', 'KB', 'MB', 'GB', 'MiB', 'GiB', 'AB', 'AiB', 'COMMADEC'],
	filterFunc: function(f, s) {
		if (!s) return 0;
		nuDump("filterfunc START "+f+" "+s);
		if (typeof(s) == 'number') s = s+"";
		nuDump("filterfunc END");
		
		// convert euro comma decimal points to dot decimal point
		if (f == 'COMMADEC')
			return s.replace(/ /g, '').replace(/,/g, '.');
		
		// treat dash as 0
		if (s == '-')
			s = '0';
		
		// axe away commas
		s = s.replace(/,/g, '');
		// we're using MB as std unit because getIntPref is 32bit only
		if (f == 'AB') {
			// auto detect if MB or GB
			if (s.indexOf("M") != -1 || s.indexOf("m") != -1)
				f = 'MB';
			else if (s.indexOf("G") != -1 || s.indexOf("g") != -1)
				f = 'GB';
			else
				f = 'B';
		}
		else if (f == 'AiB') {
			// auto detect if MB or GB
			if (s.indexOf("M") != -1 || s.indexOf("m") != -1)
				f = 'MiB';
			else if (s.indexOf("G") != -1 || s.indexOf("g") != -1)
				f = 'GiB';
			else
				f = 'B';
		}
		
		if (f == 'B')
			return Math.round(parseInt(s, 10)/1000000);
		else if (f == 'KB')
			return Math.round(parseInt(s, 10)/1000);
		else if (f == 'GB')
			return Math.round(parseFloat(s)*1000);
		else if (f == 'MB') // assume mbytes
			return parseInt(s, 10);
		else if (f == 'MiB') // assume mbytes
			return Math.round(parseInt(s, 10)*1024*1024/1000000);
		else if (f == 'GiB') // assume mbytes
			return Math.round(parseInt(s, 10)*1024*1024*1024/1000000);
	},


	// load user preferences
	loadPrefs: function () {
		var doinittime = false;
		var ee, e;
		
		try {
			var prefSvc = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
			var pref = prefSvc.getBranch("extensions.netusage."+this.mitemid+".");
			
			try {
				this.mDebug = pref.getBoolPref("debug");
			} catch(ee) { }
			try {
				this.mFetchOnStartup = pref.getIntPref("fetchonstartup");
			} catch(ee) { }
			
			var prefISP = pref.getCharPref("userisp");
			if (prefISP && prefISP != '' && prefISP != 'na') {
				if (this.isp == null || this.isp.id != prefISP)
					this.isp = this.getISPInterface(prefISP);

			} else {
				if (this.isp != null) {
					this.isp = null;
					doinittime = true; // disable the timer
				}
			}
			var prefCheckInterval = pref.getIntPref("checkinterval");
			// if timer checkInterval is less than 10 mins (minimum req, so ISP servers aren't DOS'd)
			// initTimer() will disable the timer
			if (this.checkInterval != prefCheckInterval) {
				this.checkInterval = prefCheckInterval;
				doinittime = true; // refresh timer interval
			}
			if (doinittime)
				this.initTimer();

		} catch(e) { nuDump('SOFTFAIL ' + e); }


	},


	//
	// nsIObserver
	//
	observe: function (subject, topic, data) {
		nuDump('Observed: '+subject+' :: '+topic+' ::: '+data);
		switch (topic) {
			case 'timer-callback':
				this.updateUsage(Components.interfaces.nsINetUsageItem.UPDATE_AUTO);
				break;
			case 'nsPref:changed':
				this.loadPrefs();
				break;
			case 'http-on-examine-response':
				subject.QueryInterface(Components.interfaces.nsIHttpChannel);
				// if our channel
				if (this.xmlhttp.channel == subject) {
					// nuke WWW-Authenticate if exist so ugly dialog doesn't pop up if wrong pass
					try {
					if (subject.getResponseHeader('WWW-Authenticate')) {
						subject.setResponseHeader('WWW-Authenticate', '', false);
						nuDump('WWW-Authenticate safely contained');
					}
					} catch(e) { }
					
					var hcontenttype, hlocation;
					try {
					hcontenttype = subject.getResponseHeader("Content-Type");
					} catch(e) { hcontenttype = '';}
					try {
					hlocation = ' -> '+subject.getResponseHeader("Location");
					} catch(e) { hlocation = ''; }
					this.vars['lasturl'] = subject.URI.spec;
		            this.vars['referer'] = subject.URI.spec;
					this.mDebughttp += this.step+') '+hcontenttype+' '+subject.responseStatus+' '+subject.URI.spec+hlocation+"\n";
				}
				break;
		}
	},



	//
	// nsIDOMEventListener
	//
	handleEvent: function(event) {
		nuDump("handleEvent: "+event.type);
		
		// check if we aborted
		if (this.mStatus == 'aborted') {
			//exit
			// TODO WE NEVER REACH HERE?
			nuDump('aborted');
			//this.observerSvc.notifyObservers(this, 'netusage-update-gui', 'draw');
			return;
		}
		
		try {

		if (event.type == 'load') {
			this.vars['lasttext'] = this.xmlhttp.responseText;
			// XMLHttpLoadEvent
			if (this.xmlhttp.status == 302) {
				// fix for firefox 3 b5 and bigpond!
				nuDump("302 redirect to -> " + this.xmlhttp.getResponseHeader("Location"));
				this.stepUpdateUsage(["GET",this.xmlhttp.getResponseHeader("Location")]);
			}
			else if (this.xmlhttp.status != 200 && this.xmlhttp.status != 0 /* for file:/// */ && (typeof(this.vars['ALLOW_ALL_HTTP_CODES']) == 'undefined' || !this.vars['ALLOW_ALL_HTTP_CODES'])) {
				// HTTP error occured
				this.mStatus = 'error-isp';
				if (this.xmlhttp.status == 401) {
					// authorization required
					this.mErrorText = 'Incorrect username or password.';
				} else {
					this.mErrorText = 'ISP error occured: '+this.xmlhttp.status+' '+this.xmlhttp.statusText+' (S'+this.step+')';
				}
				this.endUpdate(null, null);
			}
			else
				this.resumeProcess();
		}
		else if (event.type == 'error') {
			// XMLHttpErrorEvent
			this.endUpdate('error', 'Connection error occured (S'+this.step+')');
		}
		
		} catch(e) {
			this.endUpdate('error', 'Exception: ' + e + ' (S'+this.step+') ' + e.stack);
		}

	},


	// nsIAuthPromptProvider
	getAuthPrompt: function(aPromptReason) {
		throw Components.results.NS_ERROR_NOT_AVAILABLE;
	},

	// nsIAuthPrompt
	prompt: function(dialogTitle, text, passwordRealm, savePassword, defaultText, result) {
		return false;
	},
	promptUsernameAndPassword: function(dialogTitle, text, passwordRealm, savePassword, user, pwd) {
		return false;
	},
	promptPassword: function(dialogTitle, text, passwordRealm, savePassword, pwd) {
		return false;
	},

	// nsIProgressEventSink
	onProgress: function(aRequest, aContext, aProgress, aProgressMax) {},
	onStatus: function(aRequest, aContext, aStatus, aStatusArg) {},

	//
	// nsISupports
	//
	
	
	interfaces: [Components.interfaces.nsINetUsageItem,
		Components.interfaces.nsIObserver,
		Components.interfaces.nsIDOMEventListener,

		Components.interfaces.nsIAuthPromptProvider,
		Components.interfaces.nsIAuthPrompt,
		Components.interfaces.nsIProgressEventSink,
		
		Components.interfaces.nsIInterfaceRequestor,
		Components.interfaces.nsISupports
	],
	QueryInterface: function(iid) {
		if (!this.interfaces.some( function(v) { return iid.equals(v) } ))
			throw Components.results.NS_ERROR_NO_INTERFACE;
		return this;
	},

	getInterface: function(iid) {
		return this.QueryInterface(iid);
	},
}



function nsNetUsageItem() {

	this.xmlhttp = null;
	this.timer = null;
	this.lockedtimer = null;

	this.step = 0;
	this.vars = null;
	this.mLastFetched = 0;
	this.mStatus = "";
	this.mDebug = false;
	this.mFetchOnStartup = -1;



}

// MODULE
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function NSGetModule(compMgr, fileSpec) {
	return XPCOMUtils.generateModule([nsNetUsageItem]);
}

if (XPCOMUtils.generateNSGetFactory)
	var NSGetFactory = XPCOMUtils.generateNSGetFactory([nsNetUsageItem]);
