/*

The NetUsage Project
http://netusage.mozdev.org/

Copyright (C) 2006-2010 Loune Lam <lpgcritter at nasquan com>
This extension can be freely distributed under
the terms of the Mozilla Public Licence 1.1
See the licence at http://www.mozilla.org/MPL/

---------
Net Usage Item
XPCOM

*/


nsNetUsageSvc.prototype = {
	classDescription: "Net Usage Service",
	classID: Components.ID("{823EB6E4-7DC8-474D-B15F-DB10367662BC}"),
	contractID: "@netusage.iau5.com/netusagesvc;1",
	_xpcom_categories: [{
		category: "profile-after-change"
	},
	{
		category: "profile-change-teardown"
	}],

	get version() { return this.mVersion; },

	init: function() {
		var scriptinc = Components.classes["@mozilla.org/moz/jssubscript-loader;1"].getService().QueryInterface(Components.interfaces.mozIJSSubScriptLoader);
		scriptinc.loadSubScript("chrome://netusageitem/content/usagelib.js");

		var nusvc = this;

		nuGetExtensionVersion(function (version) { nusvc.mVersion = version });

		// create a dir for usage reports
		var file = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties).get("ProfD", Components.interfaces.nsIFile);
		file.append("netusage");
		if (!file.exists() || !file.isDirectory()) {
			//file.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0664);
		}
		
		// open welcome page for first run
		var prefSvc = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
		var pfr = 1;
		var pref;
		try {
		pref = prefSvc.getBranch("extensions.netusage.");
		pfr = pref.getIntPref("firstrun");
		} catch(e) { }
		try {
		pref = prefSvc.getBranch("extensions.netusage.");
		pref.setIntPref("firstrun", 0);
		} catch(e) { }
		
		if (pfr) {
			var ww = Components.classes["@mozilla.org/embedcomp/window-watcher;1"].getService(Components.interfaces.nsIWindowWatcher);
			ww.openWindow(ww.activeWindow, "chrome://netusageitem/content/welcome.xhtml", null, "titlebar,dialog=no,width=505,height=500", null);
		}
	},
	
	uninit: function() {
		for (var id in this.items) {
			this.items[id].QueryInterface(Components.interfaces.nsINetUsageItem);
			this.items[id].uninit();
			this.items[id] = null;
		}
	},
	
	//
	// nsINetUsageSvc
	//
	
	getItem: function(id) {
		if (!this.items[id]) {
			this.items[id] = Components.classes['@netusage.iau5.com/netusageitem;1'].createInstance(Components.interfaces.nsINetUsageItem);
			this.items[id].init(id);
		}
		
		return this.items[id];
	},
	
	acquireLock: function(name) {
		if (!this.locks[name])
			return (this.locks[name] = true);
		else
			return false;
	},
	
	releaseLock: function(name) {
		this.locks[name] = false;
	},

	//
	// nsIObserver
	//

	observe: function (subject, topic, data) {
		switch (topic) {
			case 'profile-after-change':
				// fun starts here
				this.init();
				break;
				
			case 'profile-change-teardown':
				// TODO uninit
				nuDump(topic);
				this.uninit();
				break;
		}
	},



	//
	// nsISupports
	//


	QueryInterface: function (iid) {
		if (!iid.equals(Components.interfaces.nsINetUsageSvc) &&
			!iid.equals(Components.interfaces.nsISupports) &&
			!iid.equals(Components.interfaces.nsIObserver)) {
			throw Components.results.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},
}



function nsNetUsageSvc() {
	this.items = {};
	this.locks = {};
}




// MODULE
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function NSGetModule(compMgr, fileSpec) {
	return XPCOMUtils.generateModule([nsNetUsageSvc]);
}

if (XPCOMUtils.generateNSGetFactory)
	var NSGetFactory = XPCOMUtils.generateNSGetFactory([nsNetUsageSvc]);
